@@ -5,7 +5,7 @@ use strict;
 use vars qw(
   $VERSION
   @_table
-  %cmd_codes %srv_codes
+  %cmd_codes %srv_codes %status_codes
   %_parsers %_msg_parsers %_meta_parsers
   %_builders %_msg_builders
 );
@@ -15,7 +15,7 @@ use IO::Select;
 use Time::Local;
 use Math::BigInt;
 
-$VERSION = '0.12';
+$VERSION = '0.15';
 
 
 # "encryption" table (grumble grumble...)
@@ -123,6 +123,24 @@ $VERSION = '0.12';
 );
 
 
+
+%status_codes = (
+  ONLINE                  => 0x0,
+  AWAY                    => 0x1,
+  DO_NOT_DISTURB_2        => 0x2,
+  NOT_AVAILABLE           => 0x4,
+  NOT_AVAILABLE_2         => 0x5,
+  OCCUPIED                => 0x10,
+  DO_NOT_DISTURB          => 0x13,
+  FREE_FOR_CHAT           => 0x20,
+  INVISIBLE               => 0x100,
+  WEB_AWARE               => 0x10000,
+  SHOW_IP                 => 0x20000,
+  TCP_MUST_AUTH           => 0x10000000,
+  TCP_IF_ON_CONNECTLIST   => 0x20000000
+);
+
+
 =head1 NAME
 
 Net::ICQ - Perl interface to an ICQ server
@@ -207,6 +225,7 @@ sub new {
     _handlers => {},
     _last_keepalive => undef,
     _connected => 1,
+    _seen_seq => [],
     _debug => 0
   };
 
@@ -325,7 +344,8 @@ sub add_handler {
     $srv_codes{$command} :
     $command;
 
-  print "=== add handler <", sprintf("%04X", $command_num), "> = $sub\n";
+  print "=== add handler <", sprintf("%04X", $command_num), "> = $sub\n"
+      if $self->{_debug};
 
   $self->{_handlers}{$command_num} = $sub;
 }
@@ -515,11 +535,18 @@ message 'Hello world' to UIN 1234:
     $offset = 4;
     foreach ('nickname', 'firstname', 'lastname', 'email') {
       $length                 = _bytes_to_int($event->{params}, $offset, 2);
-      $parsedevent->{$_}      = _bytes_to_str($event->{params}, $offset + 2, $length - 1);
-      $offset                += $length;
+      $offset += 2; # Fixed: NN 06 jan 01
+      $parsedevent->{$_}      = _bytes_to_str($event->{params}, $offset, $length - 1);
+      $offset += $length;
     }
     $parsedevent->{authorize} = _bytes_to_str($event->{params}, $offset, 1);
-    $event->{params}          = $parsedevent;
+    $event->{params} = $parsedevent;
+
+    # AUTHORIZE can contain either 00 or 01:
+    #   00 means that your client should request authorization before
+    #      adding this user to the contact list.
+    #   01 means that authorization is not required to add him/her to
+    #      your contact list.
   },
   # SRV_END_OF_SEARCH
   160 => sub {
@@ -540,46 +567,24 @@ message 'Hello world' to UIN 1234:
   # SRV_RECV_MESSAGE
   220 => sub {
     my ($event) = @_;
-    my ($parsedevent, @strings, @tmp);
+    my ($parsedevent, @time);
 
-    $parsedevent->{uin}    = _bytes_to_int($event->{params}, 0, 4);
-    $parsedevent->{time}   = timelocal(0,     #sec
-      _bytes_to_int($event->{params}, 9, 1),  #min
-      _bytes_to_int($event->{params}, 8, 1),  #hour
-      _bytes_to_int($event->{params}, 7, 1),  #day
-      _bytes_to_int($event->{params}, 6, 1),  #mon
-      _bytes_to_int($event->{params}, 4, 2)   #year
+    # Remove the bytes storing the time of the message, which makes the
+    # params look just like a regular online message (SRV_SYS_DELIVERED_MESS).
+    # Then, we can use that handler directly instead of copying its code here.
+    # Mirabilis really dropped the ball on this one, defining two separate
+    # events where it should really just be one...
+    @time = splice(@{$event->{params}}, 4, 6, ());
+    &{$_parsers{260}}($event);
+
+    # we still need to insert the time
+    $event->{params}->{time} = timelocal(0, # sec
+      _bytes_to_int(\@time, 5, 1),          # min
+      _bytes_to_int(\@time, 4, 1),          # hour
+      _bytes_to_int(\@time, 3, 1),          # day
+      _bytes_to_int(\@time, 2, 1)-1,        # mon (thanks Bek Oberin for the -1)
+      _bytes_to_int(\@time, 0, 2)           # year
     );
-    $parsedevent->{type}   = _bytes_to_int($event->{params}, 10, 2);
-    $parsedevent->{length} = _bytes_to_int($event->{params}, 12, 2);
-
-    @strings = _bytes_to_strlist([@{$event->{params}}[14..@{$event->{params}}-1]]);
-    if      ($parsedevent->{type} == 1) {
-      $parsedevent->{text}        = $strings[0];
-    } elsif ($parsedevent->{type} == 4) {
-      $parsedevent->{description} = $strings[0];
-      $parsedevent->{url}         = $strings[1];
-    } elsif ($parsedevent->{type} == 6) {
-      $parsedevent->{nickname}    = $strings[0];
-      $parsedevent->{firstname}   = $strings[1];
-      $parsedevent->{lastname}    = $strings[2];
-      $parsedevent->{email}       = $strings[3];
-      $parsedevent->{reason}      = $strings[4];
-    } elsif ($parsedevent->{type} == 8) {
-    } elsif ($parsedevent->{type} == 12) {
-      $parsedevent->{nickname}    = $strings[0];
-      $parsedevent->{firstname}   = $strings[1];
-      $parsedevent->{lastname}    = $strings[2];
-      $parsedevent->{email}       = $strings[3];
-    } elsif ($parsedevent->{type} == 19) {
-      $parsedevent->{contacts} = {};
-      shift @strings; # remove first element - number of contacts
-      for (my $i=0; $i<@strings-1; $i+=2) {
-	$parsedevent->{contacts}{$strings[$i]} = $strings[$i+1];
-      }
-    }
-
-    $event->{params} = $parsedevent;
   },
   # SRV_X2
   230 => sub {
@@ -619,6 +624,20 @@ message 'Hello world' to UIN 1234:
       $parsedevent->{firstname}   = $strings[1];
       $parsedevent->{lastname}    = $strings[2];
       $parsedevent->{email}       = $strings[3];
+    } elsif ($parsedevent->{type} == 13) {
+      $parsedevent->{name}        = $strings[0];
+      $parsedevent->{unknown1}    = $strings[1];
+      $parsedevent->{unknown2}    = $strings[2];
+      $parsedevent->{email}       = $strings[3];
+      $parsedevent->{unknown3}    = $strings[4]; #always has value: 3
+      $parsedevent->{message}     = $strings[5];
+    } elsif ($parsedevent->{type} == 14){
+      $parsedevent->{name}        = $strings[0];
+      $parsedevent->{unknown1}    = $strings[1];
+      $parsedevent->{unknown2}    = $strings[2];
+      $parsedevent->{email}       = $strings[3];
+      $parsedevent->{unknown3}    = $strings[4]; #always has value: 3
+      $parsedevent->{message}     = $strings[5];
     } elsif ($parsedevent->{type} == 19) {
       $parsedevent->{contacts} = {};
       shift @strings; # remove first element - number of contacts
@@ -631,11 +650,64 @@ message 'Hello world' to UIN 1234:
   },
   # SRV_INFO_REPLY
   280 => sub {
-    #FIX : don't know what to do here ..
+    # (same as SRV_USER_FOUND, above)
+    my ($event) = @_;
+    my ($parsedevent, $offset, $length);
+
+    $parsedevent->{uin}       = _bytes_to_int($event->{params}, 0, 4);
+    $offset = 4;
+    foreach ('nickname', 'firstname', 'lastname', 'email') {
+      $length                 = _bytes_to_int($event->{params}, $offset, 2);
+      $offset += 2; # Fixed: NN 06 jan 01
+      $parsedevent->{$_}      = _bytes_to_str($event->{params}, $offset, $length - 1);
+      $offset += $length;
+    }
+    $parsedevent->{authorize} = _bytes_to_str($event->{params}, $offset, 1);
+    $event->{params} = $parsedevent;
   },
   # SRV_EXT_INFO_REPLY
   290 => sub {
-    #FIX : don't know what to do here ..
+    # Thanks to Nezar Nielsen for this bit.
+    my ($event) = @_;
+    my ($parsedevent, $offset, $length);
+
+    $parsedevent->{uin}            = _bytes_to_int($event->{params}, 0, 4);
+    my $citylength                 = _bytes_to_int($event->{params}, 4, 2);
+    $parsedevent->{city}           = _bytes_to_str($event->{params}, 6, $citylength - 1);
+    $offset = 6 + $citylength;
+    $parsedevent->{country_code}   = _bytes_to_int($event->{params}, $offset, 2);
+    $offset += 2;
+    $parsedevent->{country_status} = _bytes_to_int($event->{params}, $offset,1);
+    $offset += 1;
+    my $statelength                = _bytes_to_int($event->{params}, $offset,2);
+    $offset += 2;
+    $parsedevent->{state}          = _bytes_to_str($event->{params}, $offset,$statelength - 1);
+    $offset += $statelength;
+    $parsedevent->{age}            = _bytes_to_int($event->{params}, $offset, 2);
+    $offset += 2;
+    $parsedevent->{sex}            = _bytes_to_int($event->{params}, $offset, 1);
+    $offset += 1;
+    for('phone', 'home_page', 'about'){
+       my $length                  = _bytes_to_int($event->{params}, $offset, 2);
+       $offset += 2;
+       $parsedevent->{$_}          = _bytes_to_str($event->{params}, $offset, $length - 1);
+       $offset += $length;
+    }
+    # done parsing
+    $event->{params} = $parsedevent;
+
+    # And from the specification (pretty much), here is some extra info:
+    #
+    # The code used in COUNTRY_CODE is the international telephone prefix, e.g.
+    #   01 00 (1) for the USA, 2C 00 (44) for the UK, 2E 00 (46) for Sweden, etc.
+    #   COUNTRY_STATUS is normally FE, unless the remote user has not entered a
+    #   country, in which case COUNTRY_CODE will be FF FF, and COUNTRY_STATUS
+    #   will be 9C.
+    # The field AGE has the value FF FF if the user has not entered his/her age.
+    # Values for SEX:
+    #   00 = Not specified
+    #   01 = Female
+    #   02 = Male
   },
   # SRV_STATUS_UPDATE
   420 => sub {
@@ -758,7 +830,7 @@ message 'Hello world' to UIN 1234:
       _int_to_bytes(4, 0xD5),
       _str_to_bytes($params->{client_ip}),
       _int_to_bytes(1, 4),
-      _int_to_bytes(4, 0x200),
+      _int_to_bytes(4, $status_codes{ONLINE}),
       _int_to_bytes(2, 6),
       _int_to_bytes(2, 0),
       _int_to_bytes(4, 0),
@@ -786,6 +858,7 @@ message 'Hello world' to UIN 1234:
     my ($ret, $num);
 
     $num = $params->{num_contacts};
+    # FIX: this shouldn't croak!  handle it gracefully..
     croak ("120 contact limit, send more than one packet")
       if ($num > 120);
 
@@ -798,10 +871,10 @@ message 'Hello world' to UIN 1234:
   },
   #CMD_SEARCH_UIN
   1050 => sub {
+    # thanks to Germain Malenfant for the fix
     my ($params) = @_;
     return [
-      _int_to_bytes(2, $params->{search_seq}),
-      _int_to_bytes(4, $params->{search_uin})
+      _int_to_bytes(4, $params->{uin})
     ];
   },
   #CMD_SEARCH_USER
@@ -1079,16 +1152,32 @@ sub _do_incoming {
     if ($self->{_debug}) {
       print '<-- event #', $event->{seq_num_1}, ' ';
       _print_packet(\@packet);
+      print " <", $event->{command},">\n";
     }
 
     # put acks in separate array because they will be handled immediately.
     if ( $event->{is_ack} ) {
         push @{$self->{_acks_incoming}}, $event;
     }
-    # stick everything else in the incoming events list
+    # stick everything that hasn't already been seen in the incoming events list
     else {
-        push @{$self->{_events_incoming}}, $event;
-    }
+      my $not_in_array = 1;
+      foreach my $seq ( @{$self->{_seen_seq}} ) { 
+	if ($seq == $event->{seq_num_1}) {
+	  $not_in_array = 0;
+	  last;
+	}
+      }
+      if ($not_in_array) {
+	  push @{$self->{_events_incoming}}, $event;
+	  push @{$self->{_seen_seq}}, $event->{seq_num_1};
+
+	  if (@{$self->{_seen_seq}} > 20) {
+	    shift @{$self->{_seen_seq}};
+	  }
+      } 
+      
+    } # end else
   } # end while
 } # end sub _do_incoming
 
@@ -1166,6 +1255,7 @@ sub _do_multis {
       if ($self->{_debug}) {
 	print ' <+ multi #', $event->{seq_num_1}, ' ';
 	_print_packet(\@packet);
+	print "\n";
       }
 
     } # end for
@@ -1238,7 +1328,7 @@ sub _do_handlers {
     # if a handler for this event has been registered
     if (exists $self->{_handlers}{$_->{command}} ) {
       # parse the raw event params
-      my $parsedevent = &{$_parsers{$_->{command}}}($_)
+      &{$_parsers{$_->{command}}}($_)
 	if ( exists $_parsers{$_->{command}} );
 
       #call the handler
@@ -1480,11 +1570,14 @@ sub _parse_packet {
   my ($event, @params);
 
   # sanity checks
-  # FIX: maybe handle these more gracefully?  :)
-  _bytes_to_int($packet, 3, 4) == $self->{_session_id}
-    or croak("Server told us the wrong session ID!");
-  _bytes_to_int($packet, 13, 4) == $self->{_uin}
-    or croak("Server told us the wrong UIN!");
+  if (_bytes_to_int($packet, 3, 4) != $self->{_session_id}) {
+    print("OOPS: Server told us the wrong session ID!\n") if $self->{_debug};
+    $self->{_connected} = 0;
+  }
+  if (_bytes_to_int($packet, 13, 4) != $self->{_uin}) {
+    print("OOPS: Server told us the wrong UIN!\n") if $self->{_debug};
+    $self->{_connected} = 0;
+  }
 
   # fill in the event's fields
   $event = {};
@@ -1539,7 +1632,8 @@ sub _str_to_bytes {
   my ($string, $add_zero) = @_;
   my (@ret);
 
-  foreach (split('', $string)) {
+  # the ?: keeps split() from complaining about undefined values
+  foreach (split('', defined($string) ? $string : '')) {
     push @ret, ord($_);
   }
   push @ret, 0 if $add_zero;
@@ -1579,12 +1673,13 @@ sub _bytes_to_int {
 # _bytes_to_str([0x12, 'f', 'o', 'o', '!'], 1, 3) == 'foo'
 
 sub _bytes_to_str {
+  # thanks to Dimitar Peikov for the fix
   my ($array, $start, $bytes) = @_;
   my ($ret);
 
   $ret = '';
   for (my $i = $start; $i < $start+$bytes; $i++) {
-    $ret .= (chr($array->[$i]) or '');
+    $ret .= $array->[$i] ? chr($array->[$i]) : '';
   }
 
   return $ret;
@@ -1637,7 +1732,7 @@ sub _print_packet {
   foreach (@$packet) {
     print sprintf("%02X ", $_);
   }
-  print "]\n";
+  print "]";
 
 }
 
